/*
 * Copyright 1996-2001 Hans Reiser
 */
#include "fsck.h"
#include <stdlib.h>


#define NR_TO_READ 8

static unsigned long tmp_zeroed;


/* leaves */
reiserfs_bitmap_t leaves_bitmap;
#define pass0_is_leaf(block) __is_marked (leaves, block)
#define pass0_mark_leaf(block) __mark (leaves, block)

/* nodes which are referred to from only one indirect item */
reiserfs_bitmap_t good_unfm_bitmap;
#define pass0_is_good_unfm(block) __is_marked (good_unfm, block)
#define pass0_mark_good_unfm(block) __mark (good_unfm, block)
#define pass0_unmark_good_unfm(block) __unmark (good_unfm, block)

/* nodes which are referred to from more than one indirect item */
reiserfs_bitmap_t bad_unfm_bitmap;
#define pass0_is_bad_unfm(block) __is_marked (bad_unfm, block)
#define pass0_mark_bad_unfm(block) __mark (bad_unfm, block)
#define pass0_unmark_bad_unfm(block) __unmark (bad_unfm, block)



static void make_aux_bitmaps (struct super_block * s)
{
    leaves_bitmap = reiserfs_create_bitmap (SB_BLOCK_COUNT (s));

    good_unfm_bitmap = reiserfs_create_bitmap (SB_BLOCK_COUNT (s));

    bad_unfm_bitmap = reiserfs_create_bitmap (SB_BLOCK_COUNT (s));

}


/* register block some indirect item points to */
static void register_unfm (unsigned long block)
{
    if (!pass0_is_good_unfm (block) && !pass0_is_bad_unfm (block)) {
	/* this block was not pointed by other indirect items yet */
	pass0_mark_good_unfm (block);
	return;
    }

    if (pass0_is_good_unfm (block)) {
	/* block was pointed once already, unmark it in bitmap of good
           unformatted nodes and mark in bitmap of bad pointers */
	pass0_unmark_good_unfm (block);
	pass0_mark_bad_unfm (block);
	return;
    }

    assert (pass0_is_bad_unfm (block));
}


/*##################### correct leaf on pass 0 ######################*/


/* we found two items last of which has a key not greater than the key on a
   first one. To find which of then is to be thrown away we go  */


/* 'upper' item is correct if 'upper + 2' exists and its key is greater than
   key of 'upper' */
static int upper_correct (struct buffer_head * bh, struct item_head * upper,
			  int upper_item_num)
{
    if (upper_item_num + 2 < B_NR_ITEMS (bh)) {
	if (comp_keys (&upper->ih_key, &(upper + 2)->ih_key) != -1)
	    /* item-num's item is out of order of order */
	    return 0;
	return 1;
    }
    
    /* there is no item above the "bad pair" */
    return 2;
}


/* 'lower' item is correct if 'lower - 2' exists and its key is smaller than
   key of 'lower' */
static int lower_correct (struct buffer_head * bh, struct item_head * lower,
			  int lower_item_num)
{
    if (lower_item_num - 2 >= 0) {
	if (comp_keys (&(lower - 2)->ih_key, &lower->ih_key) != -1)
	    return 0;
	return 1;
    }
    return 2;
}


/* for now: recover only "." and "..". Other improperly hashed entries get
   deleted (fimxe that)*/
static int fix_entry (reiserfs_filsys_t fs, struct buffer_head * bh,
		      struct item_head * ih, char * name, int namelen,
		      struct reiserfs_de_head * deh, 
		      int item_num, int entry_num)
{
    if (get_offset (&ih->ih_key) == DOT_OFFSET && entry_num < 2) {
	/* either '.' or '..' is corrupted */
	fsck_log ("pass0: block %lu: %d-th name \"%s\" in directory %k "
		  "is not hashed properly (fixed)\n",
		  bh->b_blocknr, entry_num, bad_name (name, namelen), &ih->ih_key);
	name[0] = '.';
	if (entry_num == 0) {
	    deh->deh_offset = cpu_to_le32 (DOT_OFFSET);
	} else {
	    name[1] = '.';
	    deh->deh_offset = cpu_to_le32 (DOT_DOT_OFFSET);
	}
	return 0;
    }
    fsck_log ("pass0: block %lu: %d-th name \"%s\" in directory %k "
	      "is not hashed properly (will be removed)\n",
	      bh->b_blocknr, entry_num, bad_name (name, namelen), &ih->ih_key);

    /*cut_entry (fs, bh, item_num, entry_num);*/
    return 1;
}


/* returns one if could not recover directory item. For now it only recovers
   directory item if locations are broken. FIXME: check how that will work on
   decently broken directory entry */
static int correct_direntry (reiserfs_filsys_t fs, struct buffer_head * bh,
			     struct item_head * ih, int item_num)
{
    int entry_num;
    char * name;
    int namelen;
    struct reiserfs_de_head * deh = (struct reiserfs_de_head *)B_I_PITEM (bh, ih);
    struct reiserfs_de_head * deh1;
    __u32 prev_offset = 0;
    __u16 prev_location = ih_item_len (ih);
    int min_entry_size = 1;/* we have no way to understand whether the
                              filesystem were created in 3.6 format or
                              converted to it. So, we assume that minimal name
                              length is 1 */
    int dirty;
    __u16 proper_state;
    int minimal_location;

    if (ih_item_len (ih) / (DEH_SIZE + min_entry_size) < ih_entry_count (ih))
	/* entry count is too big */
	return 1;
    
    deh1 = deh;
    dirty = 0;
    proper_state = 0;
    set_bit (DEH_Visible, &proper_state);    
    minimal_location = DEH_SIZE * ih_entry_count (ih);

    /* in this loop we only mark entries fro deletion - fixme: this is
       probably does not work when two or more neighboring entries are broken */
    for (entry_num = 0; entry_num < ih_entry_count (ih); entry_num ++, deh ++) {

	/* make sure that entries are sorted properly */
	if (deh_offset (deh) <= prev_offset) {
	    /* fixme: we do not recover from this yet */
	    return 1;
	}
	prev_offset = deh_offset (deh);


	/* check location of the current entry */
	if (deh_location (deh) >= prev_location ||
	    deh_location (deh) < minimal_location) {
	    /* location looks suspicious, skip the name */
	    fsck_log ("%d-th entry has bad location %d\n", entry_num, deh_location (deh));
	    deh->deh_location = 0;
	    prev_location = 0xffff;
	    continue;
	}

	if (prev_location == 0xffff) {
	    /* we do not know where the name ends. Find that */
	}


	if (prev_location == 0xffff) {
	    int i, bound;
	    struct reiserfs_de_head * tmp;
	    /* previous name was skipped, so we do not know where current name
               ends. Try to find it */

	    bound = ih_item_len (ih);
	    for (tmp = deh - 1; tmp >= deh1; tmp --)
		if (deh_location (tmp)) {
		    bound = deh_location (tmp) - deh_location (deh);
		    break;
		}

	    /* fixme: we should try only down to the nearest valid entry */
	    name = name_in_entry (deh, entry_num);
	    for (i = 1; i < bound; i ++) {
		if (is_properly_hashed (fs, name, i, deh_offset (deh)))
		    break;
	    }
	    if (i == bound) {
		/* ok, make deh_locations such that directory will be looking
                   like a directory */
		struct reiserfs_de_head * tmp;

		fsck_log ("pass0: block %lu: could not find the end of %d-th name in item %H\n",
			  bh->b_blocknr, entry_num, ih);
		mark_de_bad (deh);
		/*deh->deh_location = 0;
		  prev_location = 0xffff;*/

		/* set deh_location to the nearest dehs with zeroed locations */
		for (tmp = deh - 1; tmp >= deh1; tmp --) {
		    if (deh_location (tmp)) {
			break;
		    }
		    tmp->deh_location = (tmp + 1)->deh_location + 1;
		}
		
		prev_location = deh_location (deh);
		continue;		
	    }

	    /* ok, set deh_location of previous name */
	    if (!entry_num)
		die ("correct_direntry: wrong prev location");
	    fsck_log ("correct_direntry: ok, found the end of %d-th name \"%s\"\n",
		      entry_num, bad_name (name, i));
	    (deh - 1)->deh_location = cpu_to_le16 (deh_location (deh) + i);

	    /* set deh_location to the nearest dehs with zeroed locations */
	    for (tmp = deh - 2; tmp >= deh1; tmp --) {
		if (deh_location (tmp)) {
		    break;
		}
		tmp->deh_location = (tmp + 1)->deh_location + 1;
	    }
	    /* ok, things restored */
	    prev_location = deh_location (deh);
	    continue;
	}


	prev_location = deh_location (deh);

	/* fix state */
	if (deh_state (deh) != proper_state) {
	    deh->deh_state = cpu_to_le16 (proper_state);
	    dirty = 1;
	}

	namelen = name_length (ih, deh, entry_num);
	name = name_in_entry (deh, entry_num);
	if (!is_properly_hashed (fs, name, namelen, deh_offset (deh))) {
	    /* correct "." and ".." and delele other corrupted name */
	    if (fix_entry (fs, bh, ih, name, namelen, deh, item_num, entry_num))
		mark_de_bad (deh);
	    else
		/* '.' or '..' recovered */;
	}
    }

    /* delete all entries which has been marked unrecoverable */
 start_again:
    deh = B_I_DEH (bh, ih);
    for (entry_num = 0; entry_num < ih_entry_count (ih); entry_num ++, deh ++) {
	if (de_bad (deh)) {
	    fsck_log ("correct_direntry: block %lu, item %H, %d-th entry removed\n",
		      bh->b_blocknr, ih, entry_num);
	    cut_entry (fs, bh, item_num, entry_num);
	    goto start_again;
	}
    }

    if (dirty)
	mark_buffer_dirty (bh);
#if 0
    if (is_bad_directory (ih, B_I_PITEM (bh, ih), fs->s_dev, fs->s_blocksize)) {
	die ("directory item was not recovered");
    }
#endif
    return 0;
}


/* do this on pass 0 with every leaf marked used */

/* FIXME: we can improve fixing of broken keys: we can ssfe direct items which
   go after stat data and have broken keys */
static void handle_leaf (reiserfs_filsys_t fs, struct buffer_head * bh)
{
    int i, j;
    struct item_head * ih;
    __u32 * ind_item;
    unsigned long unfm_ptr;
    int dirty = 0;

 start_again:

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	/* sometimes stat datas get k_objectid==0 or k_dir_id==0 */
	if (ih->ih_key.k_dir_id == 0 || ih->ih_key.k_objectid == 0) {
	    if (i == (B_NR_ITEMS (bh) - 1)) {
		if (i == 0) {
		    fsck_log ("pass0: item %k is alone in the block %lu\n",
			      &ih->ih_key, bh->b_blocknr);
		    return;
		}
		/* delete last item */
		delete_item (fs, bh, i - 1);
		dirty = 1;
		return;
	    }
	    /* there is next item: if it is not stat data - take its k_dir_id
               and k_objectid. if key order will be still wrong - the changed
               item will be deleted */
	    if (!is_stat_data_ih (ih + 1)) {
		ih->ih_key.k_dir_id = (ih + 1)->ih_key.k_dir_id;
		ih->ih_key.k_objectid = (ih + 1)->ih_key.k_objectid;
		set_offset (KEY_FORMAT_1, &ih->ih_key, 0);
		set_type (KEY_FORMAT_1, &ih->ih_key, TYPE_STAT_DATA);
		fsck_log ("pass0: block %lu: item %d had zero k_dir_id or k_objectid - fixed to %k\n",
			  bh->b_blocknr, i, &ih->ih_key);
		dirty = 1;
	    }
	}

	if (i &&  comp_keys (&(ih - 1)->ih_key, &ih->ih_key) != -1) {
	    /* previous item has key not smaller than the key of currect item */
	    if (is_stat_data_ih (ih - 1) && !is_stat_data_ih (ih)) {
		/* fix key of stat data such as if it was stat data of that item */
		fsck_log ("pass0: block %lu: %d-th item %k is out of order, made a stat data of %d-th (%k)\n",
			  bh->b_blocknr, i - 1, &(ih - 1)->ih_key, i, &ih->ih_key);
		(ih - 1)->ih_key.k_dir_id = ih->ih_key.k_dir_id;
		(ih - 1)->ih_key.k_objectid = ih->ih_key.k_objectid;
		set_offset (KEY_FORMAT_1, &(ih - 1)->ih_key, 0);
		set_type (KEY_FORMAT_1, &(ih - 1)->ih_key, TYPE_STAT_DATA);
		dirty = 1;
	    } else {
		/* ok, we have to delete one of these two - decide which one */
		int retval;

		/* something will be deleted */
		dirty = 1;
		retval = upper_correct (bh, ih - 1, i - 1);
		switch (retval) {
		case 0:
		    /* delete upper item */
		    fsck_log ("pass0: block %lu: %d-th (upper) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i - 1, &(ih - 1)->ih_key);
		    delete_item (fs, bh, i - 1);
		    goto start_again;

		case 1:
		    /* delete lower item */
		    fsck_log ("pass0: block %lu: %d-th (lower) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i, &ih->ih_key);
		    delete_item (fs, bh, i);
		    goto start_again;

		default:
		    /* upper item was the first item of a node */
		}

		retval = lower_correct (bh, ih, i);
		switch (retval) {
		case 0:
		    /* delete lower item */
		    fsck_log ("pass0: block %lu: %d-th (lower) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i, &ih->ih_key);
		    delete_item (fs, bh, i);
		    goto start_again;

		case 1:
		    /* delete upper item */
		    fsck_log ("pass0: block %lu: %d-th (upper) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i - 1, &(ih - 1)->ih_key);
		    delete_item (fs, bh, i - 1);
		    goto start_again;

		default:
		    /* there wer only two items in a node, so we could not
                       decide what to delete, go and ask user */
		}
		fsck_log ("pass0: which of these items looks better (other will be deleted)?\n"
			  "%h\n%h\n", ih - 1, ih);
		if (fsck_user_confirmed (fs, "1 or 2?", "1\n", 1))
		    delete_item (fs, bh, i - 1);
		else
		    delete_item (fs, bh, i);
		goto start_again;
	    }
	}

	if (is_direntry_ih (ih)) {
	    correct_direntry (fs, bh, ih, i);
	    continue;
	}


	/* FIXME: should we do something with other items? */
	if (!is_indirect_ih (ih))
	    continue;
	
	ind_item = (__u32 *)B_I_PITEM (bh, ih);
	for (j = 0; j < I_UNFM_NUM (ih); j ++) {
	    unfm_ptr = le32_to_cpu (ind_item [j]);
	    if (!unfm_ptr)
		continue;
	    
	    if (fsck_mode (fs) == FSCK_ZERO_FILES) {
		ind_item [j] = 0;
		reiserfs_bitmap_clear_bit (fsck_new_bitmap(fs), unfm_ptr);
		tmp_zeroed ++;
		dirty = 1;
		continue;
	    }

	    if (not_data_block (fs, unfm_ptr) || /* journal area or bitmap or super block */
		unfm_ptr >= SB_BLOCK_COUNT (fs)) {/* garbage in pointer */

		stat_wrong_pointer_found (fs);
		fsck_log ("pass0: %d-th pointer (%lu) in item %k (leaf block %lu) is wrong\n",
			  j, unfm_ptr, &ih->ih_key, bh->b_blocknr);
		continue;
	    }
	    if (!was_block_used (unfm_ptr)) {
	      /* this will get to a pool of allocable blocks */
	      ind_item [j] = 0;
	      dirty = 1;
	      stat_wrong_pointer_found (fs);
	      continue;
	    }

	    /* mark block in bitmaps of unformatted nodes */
	    register_unfm (unfm_ptr);
	}
    }

    if (B_NR_ITEMS (bh) < 1)
	die ("pass0: block %lu got all items deleted\n",
	     bh->b_blocknr);

    pass0_mark_leaf (bh->b_blocknr);

    if (dirty) {
	stat_pass0_leaf_corected (fs);
	mark_buffer_dirty (bh);
    }
}


static int at_least_one_used (struct super_block * s, unsigned long first, int count)
{
    int i;

    for (i = 0; i < count; i ++)
	if (was_block_used (first + i))
	    return 1;
    return 0;
}

static void go_through (struct super_block * s)
{
    struct buffer_head * bbh, * bh;
    int nr_to_read = NR_TO_READ;
    int i, j;
    char * data;
    int what_node;
    unsigned long handled_blocks = 0;
    /*FILE * block_list;*/

    /* for debug only */
    /*block_list = fopen ("pass0-leaves", "w");*/

    for (i = 0; i < SB_BLOCK_COUNT (s); i += nr_to_read) {
	if (at_least_one_used (s, i, nr_to_read)) {
	    /* at least one of nr_to_read blocks is to be checked */
	    bbh = bread (s->s_dev, i / nr_to_read, s->s_blocksize * nr_to_read);
	    if (bbh) {
		for (j = 0; j < nr_to_read; j ++) {
		    if (!was_block_used (i + j)) {
			stat_free_block_found (s);
			continue;
		    }
		    if (not_data_block (s, i + j)) {
			stat_not_data_block_found (s);
			continue;
		    }

		    stat_used_block_found (s);

		    print_how_far (&handled_blocks, g_blocks_to_read);
		    data = bbh->b_data + j * s->s_blocksize;

		    what_node = who_is_this (data, s->s_blocksize);
		    if ( what_node != THE_LEAF )
			continue;
		    
		    /* fprintf (block_list, "leaf %d\n", i + j);*/

		    /* the node looks like a leaf, but it still can be
		       not perfect */
		    bh = make_buffer (s->s_dev, i + j, s->s_blocksize, data);
		    handle_leaf (s, bh);
		    brelse (bh);
		}
		bforget (bbh);
	    } else {
		/* bread failed */
		if (nr_to_read != 1) {
		    /* we tryied to read bunch of blocks. Try to read them by one */
		    i -= nr_to_read;
		    nr_to_read = 1;
		}
	    }
	}

	if (nr_to_read == 1 && ((i + 1) % NR_TO_READ) == 0) {
	    /* we have read NR_TO_READ blocks one at time, switch back to
               reading NR_TO_READ blocks at time */
	    i -= (NR_TO_READ - 1);
	    nr_to_read = NR_TO_READ;
	}
    }
    if (i != SB_BLOCK_COUNT (s))
	die ("go_through: block counting failed");

    /*    fclose (block_list);*/
}


static void check_bitmaps (struct super_block * s)
{
    int i;
 
    stats (s)->all_blocks = SB_BLOCK_COUNT (s);

    /* find how many leaves are not pointed by any indirect items */
    for (i = 0; i < SB_BLOCK_COUNT (s); i ++) {
	if (not_data_block (s, i))
	    continue;

	if (!was_block_used (i)) {
	    /* marked free in the on-disk bitmap */
	    make_allocable (i);
	    continue;
	}
	  
	if (pass0_is_leaf (i)) {
	    /* marked in the bitmap of leaves */
	    stat_leaf_found (s);
	    if (pass0_is_good_unfm (i) || pass0_is_bad_unfm (i))
		stat_pointed_leaf_found (s);
	}

	if (pass0_is_good_unfm (i) && pass0_is_bad_unfm (i))
	    die ("check_bitmap: bad and good unformatted");

	if (pass0_is_good_unfm (i)) {
	    stat_pointed_found (s);
	    stat_pointed_once_found (s);
	    continue;
	}
	if (pass0_is_bad_unfm (i)) {
	    stat_pointed_found (s);
	    stat_pointed_more_than_once_found (s);
	    continue;
	}

	/* blocks which marked used but are not leaves and are not
	   pointed (internals in short) get into bitmap of allocable
	   blocks */
	if (!(pass0_is_leaf (i))) {
	    make_allocable (i);
	}
    }
}


static int should_skip_pass0 (struct super_block * s)
{
    struct stat st;
    unsigned long size;
    int answer;
    
    size = ((SB_BLOCK_COUNT (fs) + 7) / 8) * 4;
    if (stat (".pass0.bmps", &st) != 0 || st.st_size != size)
	return 0;
    
    answer = user_confirmed ("There seems to be correct results of pass 0 - should I skip it (Yes)?",
			     "Yes\n");
    return answer;
	
}

static void save_pass0_result (struct super_block * s)
{
    FILE * fp;
    unsigned long size;
    
    size = ((SB_BLOCK_COUNT (fs) + 7) / 8);
 
    fp = fopen (".pass0.bmps", "w");
    if (!fp)
	return;

    if (fwrite (leaves_bitmap->bm_map, size, 1, fp) != 1)
	return;
    if (fwrite (good_unfm_bitmap->bm_map, size, 1, fp) != 1)
	return;
    if (fwrite (bad_unfm_bitmap->bm_map, size, 1, fp) != 1)
	return;
    if (fwrite (fsck_allocable_bitmap (s)->bm_map, size, 1, fp) != 1)
	return;
    fclose (fp);
}


static void load_pass0_result (struct super_block * s)
{
    FILE * fp;
    unsigned long size;
    
    size = ((SB_BLOCK_COUNT (fs) + 7) / 8);

    fp = fopen (".pass0.bmps", "r");
    if (!fp)
	die ("load_pass0_result: could not avoid pass0");
    
    if (fread (leaves_bitmap->bm_map, size, 1, fp) != 1 ||
	fread (good_unfm_bitmap->bm_map, size, 1, fp) != 1 ||
	fread (bad_unfm_bitmap->bm_map, size, 1, fp) != 1 ||
	fread (fsck_allocable_bitmap (s)->bm_map, size, 1, fp) != 1)
	die ("load_pass0_result: could not avoid pass0");

    fclose (fp);  
}


int pass_0 (reiserfs_filsys_t fs)
{
    int i;

    if (fsck_mode (fs) == FSCK_ZERO_FILES) {
	reiserfs_bitmap_fill (fsck_new_bitmap (fs));
	fsck_progress ("Zeroing existing files - all blocks marked used\n");
    }

    make_aux_bitmaps (fs);

    i = should_skip_pass0 (fs);
    if (i) {
	load_pass0_result (fs);
    } else {
	fsck_progress ("Pass 0 - ");fflush (stdout);
	go_through (fs);
	fsck_progress (" - done\n");
	save_pass0_result (fs);
    }

    /* find blocks which can be allocated */
    check_bitmaps (fs);

    if (fsck_mode (fs) == FSCK_ZERO_FILES) {
	/* flush bitmaps and exit */
	fsck_progress ("Zeroing existing files - zeroed %lu blocks\n", tmp_zeroed);
	reiserfs_flush_bitmap (fsck_new_bitmap (fs), fs);
	reiserfs_close (fs);
	exit (0);
    }


    stage_report (0, fs);
    return 0;
}


int are_there_used_leaves (unsigned long from, int count)
{
    int i;

    for (i = 0; i < count; i ++)
	if (pass0_is_leaf (from + i))
	    return 1;
    return 0;
}


int is_used_leaf (unsigned long block)
{
    return pass0_is_leaf (block);
}


int how_many_leaves_were_there (void)
{
    return stats (fs)->leaves;
}

/* these are used to correct uformatted node pointers */
int is_bad_unformatted (unsigned long block)
{
    return pass0_is_bad_unfm (block);
}

/* this is for check only. With this we make sure that all pointers we
   put into tree on pass 1 do not point to leaves (FIXME), do not
   point to journal, bitmap, etc, do not point out of fs boundary and
   are marked used in on-disk bitmap */
int still_bad_unfm_ptr_1 (unsigned long block)
{
    if (!block)
	return 0;
    if (pass0_is_leaf (block))
	return 1;
    if (pass0_is_bad_unfm (block) && !is_bad_unfm_in_tree_once (block))
	return 1;
    if (not_data_block (fs, block))
	return 1;
    if (!was_block_used (block))
	return 1;
    if (block >= stats (fs)->all_blocks)
	return 1;
    return 0;
    
}


/* pointers to data block which get into tree are checked with this */
int still_bad_unfm_ptr_2 (unsigned long block)
{
    if (!block)
	return 0;
    if (is_block_used (block))
	return 1;
    if (block >= stats (fs)->all_blocks)
	return 1;
    return 0;
}


/* these are used to allocate blocks for tree building */
int are_there_allocable_blocks (int amout_needed)
{
    if (reiserfs_bitmap_zeros (fsck_allocable_bitmap (fs)) < amout_needed) {
	int zeros = 0, i;
	
	fsck_progress ("Hmm, there are not enough allocable blocks, checking bitmap...");fflush (stdout);
	for (i = 0; i < fsck_allocable_bitmap (fs)->bm_bit_size; i ++)
	    if (!reiserfs_bitmap_test_bit (fsck_allocable_bitmap (fs), i))
		zeros ++;
	fsck_progress ("there are %d zeros, btw\n", zeros);
	return 0;
    }
    return 1;
}


unsigned long alloc_block (void)
{
    unsigned long block = 0; /* FIXME: start point could be used */

    if (reiserfs_bitmap_find_zero_bit (fsck_allocable_bitmap (fs), &block)) {
	die ("alloc_block: allocable blocks counter is wrong");
	return 0;
    }
    reiserfs_bitmap_set_bit (fsck_allocable_bitmap (fs), block);
    return block;
}


void make_allocable (unsigned long block)
{
    reiserfs_bitmap_clear_bit (fsck_allocable_bitmap (fs), block);
}


unsigned long how_many_uninsertables_were_there (void)
{
    return stats (fs)->uninsertable_leaves;
}


unsigned long how_many_items_were_saved (void)
{
    return stats (fs)->saved_on_pass1;
}
